/*
** Module   : spawn.c
** Abstract : a simple tool able to spawn a process on an existing user account.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 MAG 20131127 First version.
*/

#define _GNU_SOURCE
#include <crypt.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <shadow.h>
#include <signal.h>
#include <grp.h>
//#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#define MIN_ARGS 3
#define EXIT_SUCCESS 0

#define ERR_SPAWN_NO_ROOT_PERMISSION  -1
#define ERR_SPAWN_NO_ARGS             -2
#define ERR_SPAWN_NO_PASSWORD         -3
#define ERR_SPAWN_POSIX_OPENPT        -4
#define ERR_SPAWN_GRANPT              -5
#define ERR_SPAWN_UNLOCKPT            -6
#define ERR_SPAWN_OPEN_SLAVE          -7
#define ERR_SPAWN_IOCTL               -8
#define ERR_SPAWN_SETGID              -9
#define ERR_SPAWN_SETEGID            -10
#define ERR_SPAWN_INITGROUPS         -11
#define ERR_SPAWN_SETUID             -12
#define ERR_SPAWN_SETEUID            -13
#define ERR_SPAWN_SETSID             -14
#define ERR_SPAWN_CHDIR              -15
#define ERR_SPAWN_SETPGID            -16
#define ERR_SPAWN_FORK_LEADER        -17
#define ERR_SPAWN_FORK_COMMAND       -18
#define ERR_SPAWN_CHDIR_WD           -19

#define AUTH_OK           0
#define AUTH_ERR_USER     1
#define AUTH_ERR_PASSWORD 2
#define AUTH_ERR_ROOT     3

/* Global data */
struct passwd *pw;

/* File descriptors */
int fdm, fds;

/**
* Assign a controlling terminal to a process.
*
* @return  0 - Success
*          1 - Error
*/
int assign_pty()
{
    int result;

    /* Open a pseudo-terminal device */
    fdm = posix_openpt(O_RDWR);
    if (fdm < 0)
    {
#ifdef DEBUG
        perror("Error on posix_openpt().\n");
#endif
        return ERR_SPAWN_POSIX_OPENPT;
    }

    /* Grant access to the slave pseudo-terminal device */
    result = grantpt(fdm);
    if (result != 0)
    {
#ifdef DEBUG
        perror("Error on grantpt().\n");
#endif
        return ERR_SPAWN_GRANPT;
    }

    /*  Unlock a pseudo-terminal master/slave pair */
    result = unlockpt(fdm);
    if (result != 0)
    {
#ifdef DEBUG
        perror("Error on unlockpt().\n");
#endif
        return ERR_SPAWN_UNLOCKPT;
    }

    /* open slave pseudo-terminal device */
    fds = open(ptsname(fdm), O_RDWR);
    if (fds < 0)
    {
#ifdef DEBUG
        perror("Error on open slave device.\n");
#endif
        return ERR_SPAWN_OPEN_SLAVE;
    }

    /* Use the slave as TTY */
    if (ioctl(fds, TIOCSCTTY, (char *)1) == -1)
    {
#ifdef DEBUG
        perror("Error on ioctl(TIOCSCTTY)\n");
#endif
        return ERR_SPAWN_IOCTL;
    }

    /* Set stdin, stdout and stderr */
    dup2(STDIN_FILENO, fds); // PTY becomes standard input (0)
    dup2(STDOUT_FILENO, fds); // PTY becomes standard output (1)
    dup2(STDERR_FILENO, fds); // PTY becomes standard error (2)

    /* Cancel certain signals */
    signal(SIGCHLD,SIG_DFL); /* A child process dies */
    signal(SIGTSTP,SIG_IGN); /* Various TTY signals */
    signal(SIGTTOU,SIG_IGN);
    signal(SIGTTIN,SIG_IGN);
    signal(SIGHUP, SIG_IGN); /* Ignore hangup signal */
    signal(SIGTERM,SIG_DFL); /* Die on SIGTERM */

    return EXIT_SUCCESS;
}

/**
* Execute a command on a specified user account.
* A child process is created via fork().
* The child process uid and gid are set as the new user uid and gid.
* Create a new session for child. The child became a session leader.
* Assign a controlling terminal to session leader.
* The child process environment is set from the new user account.
* Create a child process tu execute the command.
*
* @param   command
*          Command to execute
* @param   args
*          Command arguments.
* @param   workdir
*          Process working directory.
*
* @return  Process exit status.
*/
int spawn(const char* command,
          char* const argv[],
          const char* workdir)
{
    pid_t child_pid, proc_pid;
    int status, result;

    /* Creates a child process */
    child_pid = fork();

    if (child_pid == 0)
    {
        /* Set gid */
        if (setgid(pw->pw_gid) < 0  || getgid() != pw->pw_gid)
        {
            exit(ERR_SPAWN_SETGID);
        }

        /* Set egit*/
        if (setegid(pw->pw_gid) < 0 || getegid() != pw->pw_gid)
        {
            exit(ERR_SPAWN_SETEGID);
        }

        /* Initialize the supplementary group access list */
        if (initgroups(pw->pw_name, pw->pw_gid) < 0)
        {
            exit(ERR_SPAWN_INITGROUPS);
        }

        /* Set uid */
        if (setuid(pw->pw_uid) < 0 || getuid() != pw->pw_uid)
        {
            exit(ERR_SPAWN_SETUID);
        }

        /* Set euid */
        if (seteuid(pw->pw_uid) < 0 || geteuid() != pw->pw_uid)
        {
            exit(ERR_SPAWN_SETEUID);
        }

        /* Creates a session and sets the process group ID */
        if(setsid() < 0)
        {
            exit(ERR_SPAWN_SETSID);
        }

        /* Assign controlling terminal */
        result = assign_pty();
        if (result != EXIT_SUCCESS)
        {
            exit(result);
        }

        /* Set user environment */
        char* path = strdup(getenv("PATH"));
        setenv("HOME",pw->pw_dir,1);
        setenv("USER",pw->pw_name,1);
        setenv("LOGNAME",pw->pw_name,1);
        setenv("LOGIN",pw->pw_name,1); /*< Historical; only strictly needed on AIX */
        setenv("SHELL",pw->pw_shell[0] ? pw->pw_shell : "/bin/sh",1);
        setenv("PATH",path,1);
        free(path);

        /* Change the current working directory */
        if (chdir(getenv("HOME")) < 0)
        {
            if (chdir("/") < 0)
            {
#ifdef DEBUG
                perror("chdir($HOME)\n");
#endif
                exit(ERR_SPAWN_CHDIR);
            }
        }

        /* Spawn a child process */
        proc_pid = fork();

        if (proc_pid == 0)
        {
            /* Close slave pty */
            close(fds);

            if (setpgid(0, child_pid) < 0)
            {
#ifdef DEBUG
                perror("setpgid()\n");
#endif
                exit(ERR_SPAWN_SETPGID);
            }

            /* Set process working directory */
            if (chdir(workdir))
            {
#ifdef DEBUG
                perror("chdir(workdir)\n");
#endif
                exit(ERR_SPAWN_CHDIR_WD);
            }

            /* Execute command */
            exit(execvp(command, argv));
        }
        else if (proc_pid < 0)
        {
            /* On fork() fails */
            exit(ERR_SPAWN_FORK_COMMAND);
        }

        /* Close master pty */
        close(fdm);

        /* Wait for child to exit*/
        wait(&status);

        /* Exit */
        exit(status);
    }
    else if (child_pid < 0)
    {
        /* On fork() fails */
        exit(ERR_SPAWN_FORK_LEADER);
    }

    return EXIT_SUCCESS;
}

/**
* Get user account informations including uid, gid and password.
* If the user exists check for password match.
* This is basically an authentication action.
* The user root is not allowed.
*
* @param   username
*          User name.
* @param   password
*          Use password.
*
* @return  AUTH_OK           - success.
*          AUTH_ERR_USER     - user does not exists
*          AUTH_ERR_PASSWORD - password does not match.
*          AUTH_ERR_ROOT     - user is root.
*/
int check_user (const char* username, const char* password)
{
    struct spwd *sp;
    char *encrypted, *correct;

    pw = getpwnam(username);
    endpwent();

    if (!pw) return AUTH_ERR_USER; //user doesn't really exist

    if (pw->pw_uid == 0)
    {
        return AUTH_ERR_ROOT;   // user is root
    }

    sp = getspnam(pw->pw_name);
    endspent();

    correct = sp ? sp->sp_pwdp : pw->pw_passwd;
    encrypted = crypt(password, correct);

    // bad pw=2, success=0
    return strcmp(encrypted, correct) ? AUTH_ERR_PASSWORD : AUTH_OK;
}

/**
* Check executable file attributes.
*
* @param   path
*          path to executable file.
*
* @return  0 - OK.
*          1 - No root permissions.
*/
int check_root(const char *path)
{
    struct stat pstat;
    int result;

    result = stat(path, &pstat);

    if (!result)
    {
        result = (pstat.st_uid == 0) &&
                 (pstat.st_gid == 0) &&
                 (pstat.st_mode & ALLPERMS) == 04755 ? 0 : 1;
    }

    return result;
}

/**
* Get the user password from stdin.
*
* @return  Password or NULL on error.
*/
char* password()
{
    int read;
    size_t len;
    char* pswd = NULL;

    /* Read a line form console */
    read = getline(&pswd, &len, stdin);

    if (read == -1)
    {
        return NULL;
    }
    else if (pswd[read-1] == '\n')
    {
        pswd[read-1] = 0;
    }

    return pswd;
}

/**
* Main function.
* Link option: -lm -lcrypt
* After build do the following:
*     $ sudo chown root:root spawn
*     $ sudo chmod 4755 spawn
*
* @param   argc
*          Number of command line arguments.
* @param   argv
*          Command line arguments.
*
* @return  Exit status.
*/
int main(int argc, char *argv[])
{
    char *psw;
    int result;

    /* Check for root permissions */
    if (check_root(argv[0]))
    {
        puts("No root access permissions.");
        exit(ERR_SPAWN_NO_ROOT_PERMISSION);
    }

    /* Print a short help */
    if (argc < MIN_ARGS)
    {
        printf("Usage: %s <user> <workdir> <command> [args]\n", argv[0]);
        exit(ERR_SPAWN_NO_ARGS);
    }

    /* Ask for password */
    psw = password();

    if(psw == NULL)
    {
        puts("No password.");
        exit(ERR_SPAWN_NO_PASSWORD);
    }

    /* Check user account credentials */
    result = check_user(argv[1], psw);

    /* Password no longer needed. */
    free(psw);

    switch(result)
    {
        case AUTH_OK:
            result = spawn(argv[MIN_ARGS], &argv[MIN_ARGS], argv[MIN_ARGS-1]);
            break;
        case AUTH_ERR_USER:
            puts("Invalid user name.");
            break;
        case AUTH_ERR_PASSWORD:
            puts("Invalid password.");
            break;
        case AUTH_ERR_ROOT:
            puts("User root not allowed.");
            break;
    }

    exit(result);
}
